
'use strict';

const {
  getValuesFromProgram,
  getFieldsFromOptions,
  dieWithoutOneOf,
  dieOnMultiple
} = require('../lib/utils');

const { Task, Project, Sprint, Note } = require('../lib/getModels')();

const modelMap = new Map([
  ['task', Task],
  ['note', Note],
  ['sprint', Sprint],
  ['project', Project]
]);

const options = [
  {
    opt: '-n, --note <number>',
    msg: 'remove a note by its ID'
  },
  {
    opt: '-t, --task <number>',
    msg: 'remove a task by its ID'
  },
  {
    opt: '-s, --sprint <number>',
    msg: 'remove a sprint by its ID'
  },
  {
    opt: '-p, --project <number>',
    msg: 'remove a project by its ID'
  },
  {
    opt: '-T, --delete-tasks',
    msg: 'if removing a project or sprint, remove associated tasks'
  },
  {
    opt: '-N, --delete-notes',
    msg: 'also remove associated notes'
  },
  {
    opt: '-d, --dryrun',
    msg: 'log actions without performing anything destructive'
  }
];

async function handler(program) {
  const fields = getFieldsFromOptions(options);
  const values = getValuesFromProgram(fields, program);
  const {
    note,
    task,
    sprint,
    project,
    deleteTasks,
    deleteNotes,
    dryrun
  } = values;

  dieWithoutOneOf({ note, task, project, sprint});
  dieOnMultiple({ note, task, project, sprint });

  let items = await identifyItemsToRemove();

  if (dryrun) {
    return message();
  } else {
    await deRef();
    for (let key in items) {
      for (let item in items[key]) {
        const Model = modelMap.get(key.replace(/s$/, ''));
        const { id } = items[key][item];
        await Model.remove({ where: { id } });
      }
    }
    return message();
  }

  function dePluralize(n, str) {
    /* istanbul ignore else */
    if (n === 1) {
      return str.replace(/s$/, '');
    } else {
      return str;
    }
  }

  function message() {
    let dr = dryrun ? 'DRYRUN: ' : '';
    let suffix = dryrun ? 'ing' : 'ed';
    let ret = {};
    let lines = [];
    for (let key in items) {
      const len = items[key].length;
      if (len > 0) {
        lines.push(`${dr}Remov${suffix} ${len} ${dePluralize(len, key)}`);
      }
    }
    ret.display = lines.join('\n');
    return ret;
  }

  async function deRef() {
    if (task) {
      /* istanbul ignore else */
      if (!deleteNotes) {
        await Note.set({
          where: { taskId: task },
          values: { taskId: null }
        });
      }
    } else if (project) {
      /* istanbul ignore else */
      if (!deleteNotes) {
        await Note.set({
          where: { projectId: project },
          values: { projectId: null }
        });
      }
      /* istanbul ignore else */
      if (!deleteTasks) {
        await Task.set({
          where: { projectId: project },
          values: { projectId: null }
        });
      }
    } else if (sprint) {
      /* istanbul ignore else */
      if (!deleteNotes) {
        await Note.set({
          where: { sprintId: sprint },
          values: { sprintId: null }
        });
      }
      /* istanbul ignore else */
      if (!deleteTasks) {
        await Task.set({
          where: { sprintId: sprint },
          values: { sprintId: null }
        });
      }
    } else {
      return;
    }
  }

  async function identifyItemsToRemove() {
    /* istanbul ignore else */
    if (note) {
      return { notes: [{ id: note }] };
    } else if (task) {
      if (deleteNotes) {
        return Note.get({ where: { taskId: task } }).then((res) => {
          return {
            tasks: [{ id: task }],
            notes: res.map(n => n.id)
          };
        });
      } else {
        return {
          tasks: [{ id: task }]
        };
      }
    } else if (project) {
      let ret = {
        tasks: [],
        notes: [],
        projects: [{ id: project }]
      };
      if (deleteTasks) {
        ret.tasks = await Task.get({ where: { projectId: project } });
      }
      if (deleteNotes) {
        ret.notes = await Note.get({ where: { projectId: project } });
      }
      return ret;
    } else if (sprint) {
      let ret = {
        tasks: [],
        notes: [],
        sprints: [{ id: sprint }]
      };
      if (deleteTasks) {
        ret.tasks = await Task.get({ where: { sprintId: sprint } });
      }
      if (deleteNotes) {
        ret.notes = await Note.get({ where: { sprintId: sprint } });
      }
      return ret;
    }
  }
}

module.exports = { options, handler };

