'use strict';

const Redis = require('ioredis');

module.exports = init;

function init(options) {
  return new Driver(options);
}

function Driver(options) {
  this.name = 'redis';
  this.host = options.host;
  this.port = options.port;
  this.db = new Redis(options);
  this.ignoredValidationPaths = ['id'];
  this.convert = {
    Boolean: String,
    Number: String,
    Date: String
  };
}

Driver.prototype.create = async function create(options) {
  let values = Object.assign({}, options.values);
  const { type } = options;
  let lastId = await this.db.get(`latest:${type}`);
  if(!lastId) {
    await this.db.set(`latest:${type}`, 1);
    values.id = 1;
  } else {
    values.id = await this.db.incr(`latest:${type}`);
  }

  if (type === 'task') {
    await this.setTaskRefs(values);
  } else if (type === 'note') {
    await this.setNoteRefs(values);
  } else if (type === 'project') {
    await this.setProjectRefs(values);
  } else if (type === 'sprint') {
    await this.setSprintRefs(values);
  } else {
    throw new Error(`Unknown type: ${type}`);
  }

  await this.db.hmset(`${type}:${values.id}`, values);
  return this.db.hgetall(`${type}:${values.id}`);
};

Driver.prototype.get = async function get(options) {
  const queryOptions = Object.assign({}, options);
  const { where, type } = queryOptions;
  if(where && where.id) {
    return this.db.hgetall(`${type}:${where.id}`);
  } else {
    let items = await this.getAllOf(type);
    return items.filter(item => {
      let match = true;
      for (let property in where) {
        if (where[property] != item[property]) {
          match = false;
        }
      }
      return match;
    }).sort((a,b) => a.id - b.id);
  }
};

Driver.prototype.set = async function set(options) {
  const queryOptions = Object.assign({}, options);
  const { where, type, values } = queryOptions;
  if (where && where.id) {
    await this.db.hmset(`${type}:${where.id}`, values);
    return this.db.hgetall(`${type}:${where.id}`);
  } else {
    let items = await this.getAllOf(type);
    let ids = items.filter(item => {
      let match = true;
      for (let property in where) {
        if (where[property] !== item[property]) {
          match = false;
        }
      }
      return match;
    }).map(m => m.id);
    await Promise.all(ids.map(id => {
      return this.db.hmset(`${type}:${id}`, values);
    }));
    return Promise.all(ids.map(id => this.db.hgetall(`${type}:${id}`)));
  }
};

Driver.prototype.remove = async function remove(options) {
  const queryOptions = Object.assign({}, options);
  const { where, type } = queryOptions;
  let id;
  if (where && where.id) {
    id = where.id;
  } else if (where) {
    await this.get(options).then(res => id = res[0].id);
  } else {
    return new Error('remove requires a .where option');
  }
  let changes = await this.db.del(`${type}:${id}`);
  return { changes };
};

Driver.prototype.close = function close() {
  this.db.quit();
};

Driver.prototype.setProjectRefs = async function(project) {
  if (!project.completed) {
    await this.db.sadd('active:projects', project.id);
  }
};

Driver.prototype.setSprintRefs = async function(sprint) {
  if (!sprint.completed) {
    await this.db.sadd('active:sprints', sprint.id);
  }
};

Driver.prototype.setNoteRefs = async function(note) {
  if (note.projectId) {
    await this.db.sadd(`notes:project:${note.projectId}`, note.id);
  }
  if (note.sprintId) {
    await this.db.sadd(`notes:sprint:${note.sprintId}`, note.id);
  }
  if (note.taskId) {
    await this.db.sadd(`notes:task:${note.taskId}`, note.id);
  }
};

Driver.prototype.setTaskRefs = async function(task) {
  if (task.projectId) {
    await this.db.sadd(`tasks:project:${task.projectId}`, task.id);
  }
  if (task.sprintId) {
    await this.db.sadd(`tasks:sprint:${task.sprintId}`, task.id);
  }
  if (!task.completed) {
    await this.db.sadd('active:tasks', task.id);
  }
};

Driver.prototype.getAllOf = function getAllOf(type) {
  return new Promise((resolve) => {
    const scanMatches = [];
    const scanStream = this.db.scanStream({
      match: `${type}:*`
    });
    scanStream.on('data', (keys) => {
      for (let key of keys) {
        scanMatches.push(key);
      }
    });
    scanStream.on('end', () => {
      const keys = scanMatches.filter((x, i, a) => a.indexOf(x) == i);
      resolve(Promise.all(keys.map(key => this.db.hgetall(key))));
    });
  });
};
