'use strict';

const cli = require('inquirer');
const termsize = require('term-size');
let questions = [];
const chalk = require('chalk');
const { red, blue, yellow, green } = chalk;
const gUl = green.underline;
const { valueOf } = require('../lib/utils');

/* istanbul ignore next */
function log() {
  console.log.apply(console, Array.prototype.slice.call(arguments));
}

// todo: should this return an object instead?
async function getValues(choices) {
  let res = [];
  for (let choice of choices) {
    await prompt(choice, choice, true).then(rec => {
      res.push(rec[choice]);
    });
  }
  return res;
}

function say(rec) {
  console.clear();
  exports.log('created: ', rec.dataValues);
}

async function ask() {
  let q2A = questions;
  questions = [];
  let responses = await cli.prompt(q2A);
  return responses;
}

function choice(name, message, choices, go, clear) {
  questions.push({ type: 'list', name, message, choices });
  if (go) {
    if (clear) {
      console.clear();
    }
    return ask();
  }
}

function prompt(name, message, go) {
  questions.push({ type: 'input', name, message });
  if (go) {
    console.clear();
    return ask();
  }
}

function multiLineSprints(sprints, notes, tasks, separator) {
  sprints = Array.isArray(sprints) ? sprints : [sprints];
  let ret = {};
  ret.lines = sprints.map(sprint => {
    const hasTasks = !!(Array.isArray(sprint.tasks) && sprint.tasks.length > 0);
    const hasNotes = !!(Array.isArray(sprint.notes) && sprint.notes.length > 0);
    const com = sprint.completed ?
      prettyDate(valueOf(sprint.completed)) :
      'false';

    let lines = [
      `id: ${blue(sprint.id)} name: ${gUl(sprint.name)}`,
      `completed: ${blue(com)}`
    ];
    if (tasks && hasTasks) {
      lines.push('Tasks:');
      sprint.tasks.map(t => {
        lines.push(`${green(t.id)}: ${blue(t.instruction)}`);
      });
    }

    if (notes && hasNotes) {
      lines.push('Notes:');
      sprint.notes.map(n => {
        const str = [
          `${green(n.id)}: ${blue(n.text)}`,
          ` - ${red(prettyDate(n.createdAt))}`
        ].join('');
        lines.push(str);
      });
    }

    lines.push(`${separator || ''}`);
    return lines.join('\n');
  });

  return ret;
}

function multiLineProjects(projects, notes, tasks, separator, showCompleted) {
  projects = Array.isArray(projects) ? projects : [projects];
  let ret = {};
  ret.lines = projects.map(proj => {
    const hasTasks = !!(Array.isArray(proj.tasks) && proj.tasks.length > 0);
    const hasNotes = !!(Array.isArray(proj.notes) && proj.notes.length > 0);
    const due = proj.due ? prettyDate(valueOf(proj.due)) : 'none';
    const com = proj.completed ? prettyDate(valueOf(proj.completed)) : 'false';
    let lines = [
      [
        `id: ${blue(proj.id)}`,
        ` name: ${gUl(proj.name)}`,
        ` priority: ${red(proj.priority)}`
      ].join(''),
      `description: ${gUl(proj.description)}`,
      `due: ${blue(due)} completed: ${blue(com)}`
    ];
    if (tasks && hasTasks) {
      let count = showCompleted ?
        proj.tasks.length :
        proj.tasks.filter(t => !t.completed).length;
      lines.push(`Tasks(${count}):`);
      if(showCompleted) {
        proj.tasks.map(t => {
          let symbol = t.completed ? '☒':'☐';
          lines.push([
            `${red.bold(symbol)} `,
            `${green(t.instruction)}(`,
            ` ${blue.underline(t.id)} )`
          ].join(''));
        });
      } else {
        proj.tasks.map(t => {
          if(!t.completed) {
            lines.push([
              `${red.bold('☐')} `,
              `${green(t.instruction)}`,
              ` ( ${blue.underline(t.id)} )`
            ].join(''));
          }
        });
      }
    }

    if (notes && hasNotes) {
      lines.push('Notes:');
      proj.notes.map(n => {
        let id = blue(n.id);
        let text = green(n.text);
        let date = prettyDate(n.createdAt);
        lines.push(`${id}: ${text} - ${date}`);
      });
    }

    lines.push(`${separator || ''}`);
    return lines.join('\n');
  });

  return ret;
}

function multiLineNotes(notes, separator) {
  notes = Array.isArray(notes) ? notes : [notes];
  let ret = {};
  ret.lines = notes.map(note => {
    let created = blue(prettyDate(note.createdAt));
    let updated = blue(prettyDate(note.updatedAt));
    let id = blue(note.id);
    let text = gUl(wrap(note.text));
    return [
      `text: ${text}`,
      `id: ${id} created: ${created} updated: ${updated}`,
      `${getAssociation(note.taskId, note.projectId, note.sprintId)}`,
      `${separator || ''}`
    ].join('\n');
  });
  return ret;
}

function getAssociation(task, project, sprint) {
  let ret = 'Associated with';
  if (task) {
    return `${ret} task: ${task}`;
  } else if (project) {
    return `${ret} project: ${project}`;
  } else if (sprint) {
    return `${ret} sprint: ${sprint}`;
  } else {
    return `${ret}: none`;
  }
}

function wrap(str) {
  const len = 63;
  let i = 6;
  let ret = '';
  for (let char of str) {
    if (i <= len) {
      ret += char;
      i++;
    } else {
      ret += `${char}\n`;
      i = 0;
    }
  }
  return ret;
}

function multiLineTasks(tasks, showNotes, separator) {
  tasks = Array.isArray(tasks) ? tasks:[tasks];
  let ret = {};
  ret.lines = tasks.map(task => {
    const et = task.estimatedMinsToCompletion;
    const due = task.due ? prettyDate(valueOf(task.due)):'none';
    const com = task.completed ? prettyDate(valueOf(task.completed)):'false';
    const cre = prettyDate(valueOf(task.createdAt));
    const pname = task.p && task.p.name || 'none';
    const sname = task.s && task.s.name || 'none';
    let notes = task.n && task.n.length > 0;
    let noteArr;
    if (showNotes && notes) {
      noteArr = task.n.map((n) => {
        return `${green(n.text)} - ${blue(prettyDate(n.createdAt))}`;
      });
      noteArr.unshift('\n');
      notes = noteArr.join('\n');
    } else {
      notes = green(notes);
    }
    const pri = task.priority;
    const vnotes = task.verboseNotesRequired;
    return [
      `instruction: ${gUl(task.instruction)}`,
      `id: ${blue(task.id)} priority: ${red(pri)} ` +
      `due: ${blue(due)} estimatedTime: ${yellow(et)}`,
      `project: ${green(pname)} sprint: ${green(sname)} created: ${blue(cre)}`,
      `completed: ${blue(com)} requiresNotes: ${yellow(vnotes)}`,
      `notes: ${notes}`,
      `${separator || ''}`
    ].join('\n');
  });

  return ret;
}

function prettyDate(date) {
  return `${date.getMonth() + 1}/${date.getDate()}/${date.getFullYear()}`;
}

function oneLineProjects(projects, showHeader) {
  projects = Array.isArray(projects) ? projects: [projects];
  const { columns } = termsize();
  let ret = {};

  ret.header = showHeader &&
    makeHeader(projects, 'id', 'priority', 'taskCount', 'name');

  ret.lines = projects.map(makeLine);

  function makeLine(p, i) {
    let alternateColor = (i % 2 === 0);
    let s1 = showHeader ? ' '.repeat(7 - (7 - String(p.id).length - 1)) : ' ';
    let s2 = showHeader ? ' '.repeat(7 - (String(p.id).length - 2)) : ' ';
    let s3 = showHeader ? ' '.repeat(9) : ') ';
    let before = `${p.id}${s1}${p.priority}${s2}${p.taskCount}${s3}`;
    let project = shorten(before + p.name, columns);

    return colorized(project, alternateColor, showHeader);
  }

  return ret;
}

function oneLineSprints(sprints, showHeader) {
  sprints = Array.isArray(sprints) ? sprints : [sprints];
  const { columns } = termsize();
  let ret = {};

  ret.header = showHeader &&
    makeHeader(sprints, 'id', 'taskCount', 'name');

  ret.lines = sprints.map(makeLine);

  function makeLine(s, i) {
    let alternateColor = (i % 2 === 0);
    let s1 = showHeader ? ' '.repeat(8 - (8 - String(s.id).length - 1)) : ' ';
    let s2 = showHeader ? ' '.repeat(8 - (String(s.id).length - 2)) : ') ';
    let before = `${s.id}${s1}${s.taskCount}${s2}`;
    let sprint = shorten(before + s.name, columns);

    return colorized(sprint, alternateColor, showHeader);
  }

  return ret;
}

function oneLineNotes(notes, showHeader) {
  notes = Array.isArray(notes) ? notes : [notes];
  const { columns } = termsize();
  let ret = {};

  ret.header = showHeader &&
    makeHeader(notes, 'id', 'created', 'note')
      .replace(/(note)/, '         $1');

  ret.lines = notes.map(makeLine);
  return ret;

  function makeLine(n, i) {
    let alternateColor = (i % 2 === 0);
    let s1 = showHeader ? ' '.repeat(6 - (6 - String(n.id).length - 1)) : ' ';
    let s2 = showHeader ? ' '.repeat(6 - (String(n.id).length - 2)) : ') ';
    let before = `${n.id}${s1}${prettyDate(n.createdAt)}${s2}`;
    let note = shorten(before + n.text, columns);

    return colorized(note, alternateColor, showHeader);
  }
}

function oneLineTasks(tasks, showHeader) {
  tasks = Array.isArray(tasks) ? tasks : [tasks];
  const { columns } = termsize();
  let ret = {};

  ret.header = showHeader &&
    makeHeader(tasks, 'id', 'project', 'priority', 'task');

  ret.lines = tasks.map(makeLine);
  return ret;

  function makeLine(t, i) {
    let alternateColor = (i % 2 === 0);
    let s1 = showHeader ? ' '.repeat(6 - (6 - String(t.id).length - 1)) : ' ';
    let s2 = showHeader ? ' '.repeat(6 - (String(t.id).length - 2)) : ' ';
    let s3 = showHeader ? ' '.repeat(8) : ') ';
    let before = `${t.id}${s1}${t.projectId}${s2}${t.priority}${s3}`;
    let task = shorten(before + t.instruction, columns);

    return colorized(task, alternateColor, showHeader);
  }
}

function colorized(line, useAlternate, showHeader) {
  let color = useAlternate ? chalk.greenBright:chalk.green;
  if (showHeader) {
    return line
      .replace(/^([0-9]+)/, chalk.blue.bold('$1'))
      .replace(/([\s]+)([0-9/]+)/, '$1' + chalk.yellow('$2'))
      .replace(/([\s]+)([0-9]+)/, '$1' + chalk.red('$2'))
      .replace(/([\s]+)([\w\s]+)$/, '$1' + gUl('$2'));
  } else {
    let [before, after] = line.split(')');
    let [id, pro, pri] = before.split(' ');
    return [
      chalk.blue(id),
      pro ? chalk.yellow(pro):'',
      pri ? chalk.red(pri):null,
      ')',
      color.underline(after.trim())
    ].filter(l => !!l).join(' ');
  }
}

function makeHeader(arr, f1, f2, f3, f4) {
  let s1 = ' ';
  let s2 = ' ';
  const s3 = ' ';
  const longestId = arr
    .map(t => String(t.id))
    .sort((a,b) => b.length > a.length)[0].length;
  if (longestId > 6) {
    s2 = ' '.repeat(longestId - 6);
  } else {
    s1 = ' '.repeat(6 - (6 - longestId));
    s2 = ' '.repeat(6 - (6 - longestId));
  }
  return `${f1}${s1}${f2}${s2}${f3}${s3}${f4 || ''}`;
}

function shorten(str, n, newstring) {
  str = newstring || str;
  if (str.length < n) { return str; }
  return shorten(str, n, str.replace(/ ([^\s]*)$/, '...'));
}

module.exports = exports = {
  oneLineProjects,
  multiLineProjects,
  oneLineNotes,
  multiLineNotes,
  questions,
  log,
  getValues,
  say,
  ask,
  choice,
  prompt,
  cli,
  oneLineTasks,
  multiLineTasks,
  oneLineSprints,
  multiLineSprints
};
