'use strict';

const {
  dieWithout,
  dieWithoutOneOf,
  dieOnMultiple,
  idFromStringOrNum,
  fabQuery
} = require('./utils');

const { Task, Project, Sprint, Note } = require('./getModels')();

const getProjects = getFromModel(Project);
const getSprints = getFromModel(Sprint);

module.exports = {
  getNotes,
  getNote,
  getProject,
  getProjects,
  getSprint,
  getSprints,
  getTask,
  getTasks,
  setNote,
  setTask,
  setSprint,
  setProject
};

async function setProject(program) {
  const { project, name, description, priority, due } = program;
  const id = await idFromStringOrNum(project, Project);
  dieWithout({ id });
  const vals = { name, description, priority, due };
  dieWithoutOneOf(vals);
  const query = fabQuery({ id }, vals);
  return Project.set(query).then(() => {
    return Project.get({ where: { id } }).then(res => {
      return { display: JSON.stringify(res, null, 2) };
    });
  });
}

async function setSprint(program) {
  const { sprint, name, sort } = program;
  const id = await idFromStringOrNum(sprint, Sprint);
  dieWithout({id});
  dieWithoutOneOf({ name, sort });
  const query = fabQuery({ id }, { name, sort });
  return Sprint.set(query).then(() => {
    return Sprint.get({ where: { id } }).then(res => {
      return { display: JSON.stringify(res, null, 2) };
    });
  });
}

async function setTask(program) {
  const {
    task,
    sprint,
    project,
    instruction,
    verboseNotesRequired,
    priority,
    estimatedMinsToCompletion,
    due
  } = program;
  const id = await idFromStringOrNum(task, Task);
  const sprintId = await idFromStringOrNum(sprint, Sprint);
  const projectId = await idFromStringOrNum(project, Project);
  dieWithout({ id });
  const vals = {
    sprintId,
    projectId,
    instruction,
    verboseNotesRequired,
    priority,
    estimatedMinsToCompletion,
    due
  };
  dieWithoutOneOf(vals);
  const query = fabQuery({ id }, vals);

  return Task.set(query).then(() => {
    return Task.get({ where: { id } }).then(res => {
      return { display: JSON.stringify(res, null, 2) };
    });
  });
}

async function setNote(program) {
  const { text, note, sprint, task, project } = program;
  const id = await idFromStringOrNum(note, Note);
  const sprintId = await idFromStringOrNum(sprint, Sprint);
  const taskId = await idFromStringOrNum(task, Task);
  const projectId = await idFromStringOrNum(project, Project);
  dieWithout({ id });
  const vals = {
    text,
    sprintId,
    projectId,
    taskId
  };
  const query = fabQuery({ id }, vals);
  return Note.set(query).then(() => {
    return Note.get({ where: { id } }).then(res => {
      return { display: JSON.stringify(res, null, 2) };
    });
  });
}

function getNote(id) {
  return Note.get({ where: { id } });
}

async function getNotes(program) {
  const taskId = await idFromStringOrNum(program.task, Task);
  const projectId = await idFromStringOrNum(program.project, Project);
  const sprintId = await idFromStringOrNum(program.sprint, Sprint);

  dieOnMultiple({ taskId, projectId, sprintId });
  let results;

  if (taskId) {
    results = await Note.get({ where: { taskId } });
  } else if (projectId) {
    results = await Note.get({ where: { projectId } });
  }else if (sprintId) {
    results = await Note.get({ where: { sprintId } });
  }else {
    results = await Note.get();
  }

  return results;
}

function getProject(id) {
  return Project.get({ where: { id } });
}

function getFromModel(Model) {
  return async function(program) {
    program = program || {};
    let name = program.name;
    const showCompleted = program.showCompleted;
    let res;
    const query = { where: {} };

    if (name) {
      query.where.name = { like: `%${name}%` };
    }

    if (!showCompleted) {
      query.where.completed = null;
    }

    res = await Model.get(query);
    return res;
  };
}

function getSprint(id) {
  return Sprint.get({ where: { id } });
}

function getTask(id) {
  return Task.get({ where: { id } });
}

async function getTasks(program) {
  const { task, project, sprint } = program;
  const projectId = await idFromStringOrNum(project, Project);
  const sprintId = await idFromStringOrNum(sprint, Sprint);
  const query = { where: {} };
  if (!program.showCompleted) {
    query.where.completed = null;
  }

  if (task) {
    query.where.instruction = { like: task };
  } else if (projectId) {
    query.where.projectId = projectId;
  } else if (sprintId) {
    query.where.sprintId = sprintId;
  }

  return Task.get(query);
}
