'use strict';

const { isa, convert } = require('./utils');

module.exports = Schema;

function Schema(schema, ignored, converted) {
  this.seed = Object.assign({}, schema);
  validateSeed.call(this);
  this.paths = Object.keys(this.seed);
  this.virtualPaths = nonVirtualKeys.call(this);
  this.ignoredPaths = Array.isArray(ignored) ? [].concat(ignored): [];
  this.convert = Object.assign({}, converted);
  this.setRequiredFields = setRequiredFields.bind(this);
  this.castIn = castIn.bind(this);
  this.castOut = castOut.bind(this);
  this.validate = validate.bind(this);
  this.setDefaults = setDefaults.bind(this);
  this.getPathType = getPathType.bind(this);
  this.setRequiredFields();
}

function getPathType(path) {
  return this.seed[path] && this.seed[path].type;
}

function setRequiredFields() {
  this.requiredFields = [];
  for (let key in this.seed) {
    let path = this.seed[key];
    if (path.allowNull === false) {
      this.requiredFields.push(key);
    }
  }
}

// todo: refactor the hell out of this ugly shit
function validate(obj) {
  let error;
  let ignored = this.ignoredPaths;
  let failed = this.requiredFields.map(f => {
    if(obj[f] == undefined && !(ignored && ignored.includes(f))) {
      return f;
    }
  }).filter(undefs);
  error = dieOnAny(failed, 'missing required field(s)');
  if (error) { return error; }
  failed = this.paths.map(path => {
    if (!this.seed[path].enum || obj[path] == undefined) {
      return;
    }
    return this.seed[path].enum.includes(obj[path]) ? undefined : path;
  }).filter(undefs);
  error = dieOnAny(failed, 'enum match failed');
  if (error) { return error; }

  failed = this.paths.map(path => {
    if (this.seed[path].type !== Number || obj[path] == undefined) {
      return;
    }
    let { min, max } = this.seed[path];
    if (min && max) {
      return obj[path] >= min && obj[path] <= max ? undefined : path;
    } else if (min) {
      return obj[path] >= min ? undefined : path;
    } else if (max) {
      return obj[path] <= max ? undefined : path;
    } else {
      return undefined;
    }
  }).filter(undefs);
  error = dieOnAny(failed, 'Number outside of min/max range');
  if (error) { return error; }

  failed = this.paths.map(path => {
    if (this.seed[path].type !== String || obj[path] == undefined) {
      return;
    }
    let { minLength, maxLength } = this.seed[path];
    let len = obj[path].length;
    if (minLength && maxLength) {
      return len >= minLength && len <= maxLength ? undefined : path;
    }
    if (minLength) {
      return len >= minLength ? undefined : path;
    }
    if (maxLength) {
      return len <= maxLength ? undefined : path;
    }
    return undefined;
  }).filter(undefs);
  error = dieOnAny(failed, 'String length outside min/max Length');
  if (error) { return error; }

  return obj;

  function undefs(f) {
    return f !== undefined;
  }

  function dieOnAny(arr, str) {
    if (arr.length) {
      return new Error(`${str} ${arr.join(',')}`);
    }
  }
}

function castIn(obj) {
  let ret = {};
  for (let key in obj) {
    if (!this.virtualPaths.includes(key)) {
      let type = this.getPathType(key);
      if (!type) {
        continue;
      }
      if (obj[key] === null) {
        ret[key] = null;
        continue;
      }
      let typeName = type && type.name;
      let convName = this.convert[typeName] && this.convert[typeName].name;
      if (this.convert[typeName] && !convName) {
        ret[key] = this.convert[typeName](obj[key]);
        continue;
      }

      if (convName && !isa(convName, obj[key])) {
        ret[key] = convert(obj[key], typeName, convName);
      } else {
        ret[key] = obj[key];
      }
    }
  }
  return ret;
}

function castOut(obj) {
  if (Array.isArray(obj)) {
    return obj.map(castOut, this);
  }
  let ret = {};
  for (let key in obj) {
    let type = this.getPathType(key);
    let typeName = type.name;

    if (!isa(typeName, obj[key]) && obj[key]) {
      ret[key] = convert(obj[key], typeof obj[key], typeName);
    } else {
      ret[key] = obj[key];
    }
  }
  return ret;
}

function nonVirtualKeys() {
  let virtuals = [];
  for (let key in this.seed) {
    if (this.seed[key].type.name === 'Object') {
      virtuals.push(key);
    }
  }
  return virtuals;
}

function validateSeed() {
  for (let key in this.seed) {
    const path = this.seed[key];
    if (!path.type || typeof path.type !== 'function') {
      throw new Error(`${key} path requires a Function type declaration.`);
    }
  }
}

function setDefaults(obj) {
  let ret = {};
  for (let path in this.seed) {
    let { defaultValue } = this.seed[path];
    if (!obj[path] && defaultValue !== undefined) {
      ret[path] = typeof defaultValue === 'function' ?
        defaultValue() : defaultValue;
    }
  }

  return Object.assign(ret, obj);
}
