'use strict';

module.exports = {
  sql,
  redis,
  throwWithout,
  stringToBoolean,
  isa,
  convert,
  valueOf,
  getValuesFromProgram,
  getFieldsFromOptions,
  idFromStringOrNum,
  dieWithout,
  dieOnMultiple,
  dieWithoutOneOf,
  sortFunc,
  fabQuery,
  truthyOrNull
};

function truthyOrNull(obj, ignore) {
  ignore = ignore || [];
  let ret = {};
  for (let key of Object.keys(obj)) {
    if (obj[key] || obj[key] === null || ignore.includes(key)) {
      ret[key] = obj[key];
    }
  }
  return ret;
}

function fabQuery(w, v) {
  let query = { where: {}, values: {} };
  for (let key of Object.keys(w)) {
    query.where[key] = w[key];
  }
  for (let key of Object.keys(v)) {
    /* istanbul ignore else */
    if (v[key] !== undefined) {
      query.values[key] = v[key];
    }
  }
  return query;
}

function dieWithoutOneOf(object) {
  const keys = Object.keys(object);
  let die = keys
    .map(key => {
      return !!object[key];
    })
    .filter(b => b === true).length < 1;
  /* istanbul ignore else */
  if (die) {
    throw new Error(`At least One of ${keys} is required.`);
  }
}

// todo: make this work with an object or with multiple value args
function dieOnMultiple(object) {
  const keys = Object.keys(object);
  let die = keys
    .map(key => { return !!object[key]; })
    .filter(b => b === true).length > 1;
  if (die) { throw new Error(`Do Not Use ${keys} Simultaneously.`); }
}

function dieWithout(object, customError) {
  const keys = Object.keys(object);
  if (keys.length === 0) {
    throw new Error(customError || 'no keys found');
  }
  keys.map(key => {
    if ((!object[key] || typeof object[key] === 'function')) {
      throw new Error(customError || `${key} is a required param.`);
    }
  });
}

function idFromStringOrNum(val, Model) {
  return new Promise((res) => {
    if (val === undefined || typeof val === 'function') { return res(); }
    if (typeof val === 'number') { return res(val); }
    if (/^[0-9]+$/.test(val)) { return res(Number(val)); }
    if (val === '') {
      return res(null);
    }
    let query = { where: { } };
    if (Model.type === 'task') {
      query.where.instruction = { like: val };
    } else if (Model.type === 'note') {
      query.where.text = { like: val };
    } else {
      query.where.name = { like: val };
    }
    return res(Model.get(query).then(rec => {
      /* istanbul ignore else */
      if (rec && !(Array.isArray(rec))) {
        return rec.id;
      } else if (rec && Array.isArray(rec)) {
        return rec[0] && rec[0].id;
      } else {
        return;
      }
    }));
  });
}

function valueOf(val) {
  return typeof val === 'function' ? undefined : val;
}

function getValuesFromProgram(fields, program) {
  let ret = {};
  for (let field of fields) {
    ret[field] = valueOf(program[field]);
  }
  return ret;
}

function getFieldsFromOptions(options) {
  return options.map(o => {
    return o.opt
      .replace(/.*--([^\s]+).*/, '$1')
      .replace(/-([a-z])/, (_, p1) => {
        return p1.toUpperCase();
      });
  });
}

function stringToBoolean(str) {
  if (typeof str !== 'string') {
    throw new Error(`stringToBoolean expects a string, got: ${typeof str}`);
  }
  switch(str) {
    case 'true':
      return true;
    case 'false':
      return false;
    default:
      throw new Error(`stringToBoolean expects "true"|"false", got: ${str}`);
  }
}

function convert(prop, from, to) {
  switch (to) {
    case 'string':
    case 'String':
      return String(prop);
    case 'number':
    case 'Number':
      return Number(prop);
    case 'object':
    case 'Date':
      if (typeof prop === 'string' && /[^0-9]/.test(prop)) {
        return new Date(prop);
      }
      return new Date(Date.parse(prop));
    case 'boolean':
    case 'Boolean':
      if (/string/i.test(from)) {
        return stringToBoolean(prop);
      }
      return Boolean(prop);
    default:
      throw new Error(`convert expects date|string|number|boolean, got: ${to}`);
  }
}

function isa(str, prop) {
  switch (str) {
    case 'object':
    case 'Date':
      return prop instanceof Date;
    case 'string':
    case 'String':
      return typeof prop === 'string';
    case 'number':
    case 'Number':
      return typeof prop === 'number';
    case 'boolean':
    case 'Boolean':
      return typeof prop === 'boolean';
    default:
      throw new Error(`isa expects date|string|number|boolean, got: ${str}`);
  }
}

function throwWithout(obj) {
  for (let key in obj) {
    if (!obj[key]) {
      throw new Error(`${key} is required`);
    }
  }
}

function sql() {
  return {
    queryString,
    getFields,
    getBoundParameters,
    getWhere,
    getSelect,
    getUpdate,
    getOne
  };

  function getOne(options) {
    if (!options || !options.where){
      throw new Error('Must supply an options object with .where to getOne');
    }
    if (options.where && options.where.id) {
      return true;
    }
    return false;
  }

  function queryString(options) {
    const action = options.action;
    const table = options.type + 's';
    const where = getWhere(options.where);
    const fields = getFields(options.values);
    const boundParams = getBoundParameters(options.values);
    const select = getSelect(options.select);
    const update = getUpdate(options.values);

    switch(action) {
      case 'create':
        return `INSERT INTO ${table} (${fields}) VALUES (${boundParams})`;
      case 'get':
        return `SELECT ${select} FROM ${table} ${where}`;
      case 'set':
        return `UPDATE ${table} SET ${update} ${where}`;
      case 'remove':
        return `DELETE FROM ${table} ${where}`;
      default:
        throw new Error(`UNKNOWN ACTION: ${action}`);
    }
  }

  function getUpdate(obj) {
    if (!obj || Object.keys(obj).length === 0) {
      return '';
    }
    let values = [];
    for (let key in obj) {
      let val = obj[key];
      const quote = typeof val === 'string';
      val = quote ? `'${val}'` : val;
      values.push(`${key}=${val}`);
    }
    return values.join(',');
  }

  function getSelect(arr) {
    if(!arr || arr.length === 0) {
      return '*';
    }
    return arr.join(',');
  }

  function getWhere(obj) {
    if(!obj || Object.keys(obj).length === 0) {
      return '';
    }
    const conditions = [];
    for (let key in obj) {
      const isObject = !!(typeof obj[key] === 'object');
      const hasLike = !!(obj[key] && obj[key].like);
      const like = ( isObject && hasLike );
      const isNull = !!(obj[key] == null);
      const quote = typeof obj[key] === 'string';

      if (isNull) {
        conditions.push(`${key} IS NULL`);
      } else if (like) {
        conditions.push(`${key} LIKE '%${obj[key].like}%'`);
      }else {
        const q = quote ? '\'':'';
        conditions.push(`${key}=${q}${obj[key]}${q}`);
      }

    }

    return conditions.length > 1 ?
      'WHERE ' + conditions.join(' AND ') :
      'WHERE ' + conditions[0];
  }

  function getBoundParameters(obj) {
    let ret = [];
    for (let key in obj) {
      ret.push(`@${key}`);
    }
    return ret.join(',');
  }

  function getFields(obj) {
    if(!obj) {
      return;
    }
    const keys = Object.keys(obj);
    return keys.join(',').trim();
  }
}

function redis() {
  return {};
}

function sortFunc(sort, schema) {
  if(!sort) {
    return (a, b) => a.id - b.id;
  } else {
    const reversed = /^\+/.test(sort);
    sort = sort.replace(/^\+/, '');
    const type = schema[sort] ? schema[sort].type : undefined;

    if (!type) {
      throw new Error(`UnknownSortField: ${sort}`);
    }

    if (reversed) {
      return (a, b) => {
        if (type === Number || type === Date) {
          return b[sort] - a[sort];
        } else {
          // come back to why this else never fires
          /* istanbul ignore else */
          if (b[sort] < a[sort]) {
            return -1;
          } else {
            return 1;
          }
        }
      };
    } else {
      return (a, b) => {
        if (type === Number || type === Date) {
          return a[sort] - b[sort];
        } else {
          if (a[sort] > b[sort]) {
            return 1;
          } else {
            return -1;
          }
        }
      };
    }
  }
}
