#!/usr/bin/env node
'use strict';

const fs = require('fs');
const path = require('path');
const mkdirp = require('mkdirp');
const Sqlite = require('better-sqlite3');
const { red, yellow, green } = require('chalk');
const defaultConfigPath = path.join(__dirname, '../default.config.json');
const defaultConfigFile = fs.readFileSync(defaultConfigPath);
const json = JSON.parse(defaultConfigFile);
const HOME = process.env.HOME;
const TARGETDIR = `${HOME}/.gishdo`;
const CONFIGFILE = `${TARGETDIR}/config.json`;
const DBFILE = `${TARGETDIR}/gishdo.sqlite`;

json.gishdoHome = CONFIGFILE;
json.driverOptions.path = DBFILE;

setup();

async function setup() {
  log(red.underline('Starting the Setup Process.'));

  if (fs.existsSync(TARGETDIR)) {
    log(green(`${TARGETDIR} exists, leaving it alone.`));
    process.exit(0);
  } else {
    log(green(`Making ${TARGETDIR}.`));
    mkdirp(TARGETDIR, { mode: '0700' }, async (err) => {
      if (err) {
        error(err.message);
        process.exit(1);
      }
      log(green(`Making ${CONFIGFILE}`));
      fs.writeFileSync(CONFIGFILE, JSON.stringify(json, null, 2), { mode: '0700' });
      log(green('Created config file.'));
      const db = new Sqlite(DBFILE);
      await buildSqlite(db).catch(error);
      log(yellow('DB Build Complete.'));
      log(red.underline('Setup Process Complete.'));
      return db.close();
    });
  }
}

function buildSqlite(db) {
  return new Promise((resolve, reject) => {
    log(yellow(`Building the DB @ ${DBFILE}`));
    const tasks = [
      'CREATE TABLE`tasks`(`id` INTEGER PRIMARY KEY AUTOINCREMENT,',
      '`instruction` VARCHAR(128) NOT NULL, `completed` DATETIME,',
      '`minutesInvested` BIGINT, `verboseNotesRequired` TINYINT(1) DEFAULT 0,',
      '`priority` INTEGER(1) DEFAULT 4,',
      '`estimatedMinsToCompletion` INTEGER(2) DEFAULT 25,',
      '`place` INTEGER(2), `due` DATETIME, `createdAt` DATETIME NOT NULL,',
      '`updatedAt` DATETIME NOT NULL,',
      '`projectId` INTEGER REFERENCES`projects`(`id`)',
      'ON DELETE SET NULL ON UPDATE CASCADE, `sprintId` INTEGER',
      'REFERENCES`sprints`(`id`) ON DELETE SET NULL ON UPDATE CASCADE);',
      'CREATE UNIQUE INDEX`tasks_instruction_project_id`',
      'ON`tasks`(`instruction`, `projectId`);'
    ].join(' ');
    const projects = [
      'CREATE TABLE`projects`(`id` INTEGER PRIMARY KEY AUTOINCREMENT,',
      '`name` VARCHAR(24) NOT NULL, `description` VARCHAR(512) NOT NULL,',
      '`completed` DATETIME, `priority` INTEGER(1) DEFAULT 4, `due` DATETIME,',
      '`createdAt` DATETIME NOT NULL, `updatedAt` DATETIME NOT NULL);',
      'CREATE UNIQUE INDEX`projects_name` ON`projects`(`name`);'
    ].join(' ');
    const sprints = [
      'CREATE TABLE`sprints`(`id` INTEGER PRIMARY KEY AUTOINCREMENT,',
      '`name` VARCHAR(16) NOT NULL, `status` TEXT DEFAULT \'new\',',
      '`sort` TEXT, `due` DATETIME, `completed` DATETIME,',
      '`createdAt` DATETIME NOT NULL, `updatedAt` DATETIME NOT NULL);',
      'CREATE UNIQUE INDEX`sprints_name` ON`sprints`(`name`);'
    ].join(' ');
    const notes = [
      'CREATE TABLE`notes`(`id` INTEGER PRIMARY KEY AUTOINCREMENT,',
      '`text` TEXT NOT NULL, `createdAt` DATETIME NOT NULL,',
      '`updatedAt` DATETIME NOT NULL,',
      '`taskId` INTEGER REFERENCES`tasks`(`id`)',
      'ON DELETE SET NULL ON UPDATE CASCADE,',
      '`projectId` INTEGER REFERENCES`projects`(`id`)',
      'ON DELETE SET NULL ON UPDATE CASCADE, `sprintId`',
      'INTEGER REFERENCES`sprints`(`id`)',
      'ON DELETE SET NULL ON UPDATE CASCADE);'
    ].join(' ');
    try {
      db.exec(projects);
      db.exec(sprints);
      db.exec(tasks);
      db.exec(notes);
      fs.chmodSync(DBFILE, '0700');
      resolve();
    } catch(e) {
      reject(e);
    }
  });
}


function log(s) {
  return console.log(red('GISHDO: '), s);
}

function error(s) {
  return console.error(s);
}
